// Copyright (C) 2021  Nexedi SA and Contributors.
//                     Kirill Smelkov <kirr@nexedi.com>
//
// This program is free software: you can Use, Study, Modify and Redistribute
// it under the terms of the GNU General Public License version 3, or (at your
// option) any later version, as published by the Free Software Foundation.
//
// You can also Link and Combine this program with other software covered by
// the terms of any of the Free Software licenses or any of the Open Source
// Initiative approved licenses and Convey the resulting work. Corresponding
// source of such a combination shall include the source code for all other
// software used.
//
// This program is distributed WITHOUT ANY WARRANTY; without even the implied
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
// See COPYING file for full licensing terms.
// See https://www.nexedi.com/licensing for rationale and options.

package xbtree

import (
	"testing"

	"lab.nexedi.com/kirr/neo/go/zodb"
)

func TestPPTreeSubSetOps(t *testing.T) {
	const (
		a zodb.Oid = 0xa + iota
		b
		c
		d
		ø = zodb.InvalidOid
	)
	type S = PPTreeSubSet
	type testEntry struct {
		A, B       S
		Union      S
		Difference S
	}
	E := func(A, B, U, D S) testEntry {
		return testEntry{A, B, U, D}
	}

	testv := []testEntry{
		E(
			S{},	// A
			S{},	// B
				S{},	// U
				S{}),	// D

		E(
			S{a:{ø,0}},	// A
			S{a:{ø,0}},	// B
				S{a:{ø,0}},	// U
				S{}),		// D

		E(
			S{a:{ø,0}},	// A
			S{b:{ø,0}},	// B
				S{a:{ø,0}, b:{ø,0}},	// U
				S{a:{ø,0}}),		// D

		E(
			S{a:{ø,1}, b:{a,0}},	// A
			S{a:{ø,1}, c:{a,0}},	// B
				S{a:{ø,2}, b:{a,0}, c:{a,0}},	// U
				S{a:{ø,1}, b:{a,0}}),		// D

		E(
			S{a:{ø,1}, b:{a,1}, c:{b,0}},	// A
			S{a:{ø,1}, b:{a,1}, d:{b,0}},	// B
				S{a:{ø,1}, b:{a,2}, c:{b,0}, d:{b,0}},	// U
				S{a:{ø,1}, b:{a,1}, c:{b,0}}),		// D

		E(
			S{a:{ø,1}, b:{a,0}},	// A
			S{a:{ø,1}, b:{a,0}},	// B
				S{a:{ø,1}, b:{a,0}},	// U
				S{}),			// D
	}

	for _, tt := range testv {
		U := tt.A.Union(tt.B)
		D := tt.A.Difference(tt.B)

		if !U.Equal(tt.Union) {
			t.Errorf("Union:\n  A:   %s\n  B:   %s\n  ->u: %s\n  okU: %s\n", tt.A, tt.B, U, tt.Union)
		}
		if !D.Equal(tt.Difference) {
			t.Errorf("Difference:\n  A:   %s\n  B:   %s\n  ->d: %s\n  okD: %s\n", tt.A, tt.B, D, tt.Difference)
		}

		// XXX also verify U/D properties like (A+B)\B + (A+B)\A + (A^B) == (A+B) ?
	}
}
